#!/usr/bin/env python3

from keras.models import Model
from keras.layers.core import Flatten, Dense, Dropout
from keras.layers import Input, concatenate, Conv2D, MaxPooling2D, Conv2DTranspose, multiply, Lambda, \
    BatchNormalization, Activation
from keras.layers.convolutional import Convolution2D, MaxPooling2D, ZeroPadding2D
from keras.layers import Dense, Conv2D, Conv3D, MaxPooling2D, MaxPooling3D
from keras.layers import Dropout, UpSampling2D, Input, concatenate, Reshape
from keras.layers import Conv2DTranspose, Lambda, Conv3DTranspose, add, LeakyReLU
from keras.optimizers import SGD
from keras import losses
from keras.optimizers import Adam
from keras import backend as K
import random
import numpy as np
import scipy.misc
import skimage.exposure
from keras.preprocessing.image import ImageDataGenerator
import tensorflow as tf
from os.path import join, isdir
import keras.metrics
import keras.losses
from os.path import join
import os
# https://github.com/qubvel/segmentation_models


generator = 'default'

path = 'C:\\data\\crack_detection\\merged_data_patches\\'
'''
input_shape = (None, None, 3)
output_shape = (None, None, 1)
'''
#input_shape = (256, 256, 3)
input_shape = (256, 256, 1)
output_shape = (256, 256, 1)
image_extension = '.png'

batch_size = 8

depth = 5

starting_nb_layers = 64

#max_items_per_epoch = 1500

if (isdir(path)):
    folders = ['training', 'validation']
    patch_per_cat = {}
    for folder in folders:
        filenames = os.listdir(join(path, folder))
        
        # TODO!
        ids_per_cat = {'n': [], 'y': [], 'crackforest': [], 'crack500': [], 'deepcrack': [], 'imagenet': [], 'imagenetsel': [], 'determinant': [], 'flickr-tiling': []} # , 'codebrim-no': []
        for filename in filenames:
            if (filename.endswith('_X'+image_extension)):
                category = filename.split('_')[0]
                if (category not in ids_per_cat):
                    continue
                ids_per_cat[category].append(filename[:-6])
        
        for category in ids_per_cat:
            print (folder, category, len(ids_per_cat[category]))
        #print (len(ids_per_cat[False]), len(ids_per_cat[True]))
        patch_per_cat[folder] = ids_per_cat
    
def wasserstein_loss(y_true, y_pred):
    """Calculates the Wasserstein loss for a sample batch.
    The Wasserstein loss function is very simple to calculate. In a standard GAN, the
    discriminator has a sigmoid output, representing the probability that samples are
    real or generated. In Wasserstein GANs, however, the output is linear with no
    activation function! Instead of being constrained to [0, 1], the discriminator wants
    to make the distance between its output for real and generated samples as
    large as possible.
    The most natural way to achieve this is to label generated samples -1 and real
    samples 1, instead of the 0 and 1 used in normal GANs, so that multiplying the
    outputs by the labels will give you the loss immediately.
    Note that the nature of this loss means that it can be (and frequently will be)
    less than 0."""
    return K.mean(-(y_true * y_pred))


def on_epoch_end_cb(data_generator):
    folder = data_generator.folder
    cats = patch_per_cat[folder].keys()

    current_index = {}
    for cat in cats:
        np.random.shuffle(patch_per_cat[folder][cat])
        current_index[cat] = 0

    indexes = []
    has_finished = False
    while (not has_finished):
        for cat in cats:
            indexes.append(patch_per_cat[folder][cat][current_index[cat]])
            current_index[cat] += 1
            if (current_index[cat] >= len(patch_per_cat[folder][cat])):
                has_finished = True

    data_generator.indexes = indexes







smooth = 1.
def dice_coef(y_true, y_pred):
    y_true_f = K.flatten(y_true)
    y_pred_f = K.flatten(y_pred)
    intersection = K.sum(y_true_f * y_pred_f)
    return (2. * intersection + smooth) / (K.sum(y_true_f) + K.sum(y_pred_f) + smooth)


def dice_coef_loss(y_true, y_pred):
    return -dice_coef(y_true, y_pred)



keras.metrics.dice_coef = dice_coef
keras.losses.dice_coef_loss = dice_coef_loss
keras.losses.wasserstein_loss = wasserstein_loss





def loss_mean(y_true, y_pred, clip_delta=1.0):
    mask = K.cast(Lambda(lambda x: x > -1000000, output_shape=(y_pred.shape[1], y_pred.shape[2], y_pred.shape[3]))(y_true),
                  'float32')
    diff = (y_true - y_pred) * (y_true - y_pred)
    score = K.sum(diff) / (K.sum(mask) + 1)
    
    return score
    
    
def masked_mean_squared_error(y_true, y_pred):
    mask = K.cast(Lambda(lambda x: x > -10, output_shape=(y_pred.shape[1], y_pred.shape[2], y_pred.shape[3]))(y_true), 'float32')
    weighted_mask = mask * ((y_true * 4.0) + 1)
    diff = (y_true - y_pred)
    squared_difference = K.sum(diff * diff * weighted_mask)  # K.sum(((y_true - y_pred) * (y_true - y_pred)) * mask)
    squared_difference /= K.sum(weighted_mask)
    return squared_difference
    
def get_conv_normalized_layer(val, nb_layers, shape=(3, 3)):
    conv = Conv2D(nb_layers, shape, use_bias=False, padding='same')(val)
    conv = BatchNormalization()(conv)
    return Activation('relu')(conv)

def get_model():
    inputs = Input(input_shape)
    val = inputs
    # val = BatchNormalization()(val)
    nb_levels = depth
    nb_layers = starting_nb_layers

    convs = []
    for i in range(nb_levels):
        conv = Conv2D(nb_layers, (3, 3), activation='relu', padding='same')(val)
        conv = get_conv_normalized_layer(conv, nb_layers, shape=(3, 3))
        convs.append(conv)
        if (i < nb_levels - 1):
            val = MaxPooling2D(pool_size=(2, 2))(conv)
            print(val.shape)
            nb_layers *= 2
        else:
            val = conv

    for i in range(nb_levels - 1):
        nb_layers = int(nb_layers / 2)
        up = concatenate(
            [Conv2DTranspose(nb_layers, (2, 2), strides=(2, 2), padding='same')(val), convs[nb_levels - i - 2]], axis=3)
        val = Conv2D(nb_layers, (3, 3), activation='relu', padding='same')(up)
        val = get_conv_normalized_layer(val, nb_layers, (3, 3))

    val = Conv2D(output_shape[2], (1, 1), activation='sigmoid')(val)
    #val = Conv2D(output_shape[2], (1, 1), activation='linear')(val)

    # val = multiply([val, mask])

    model = Model(inputs=inputs, outputs=[val])

    model.compile(optimizer=Adam(),
                  loss=losses.binary_crossentropy)  # lr=1e-4 #dice_coef_loss   losses.mean_squared_error
    # lr=1e-3
    return model



def process_patch(im_np, is_training):
    return im_np
    
datagen = ImageDataGenerator(
    vertical_flip = True,
    horizontal_flip=True,
    rotation_range=360,
    fill_mode='constant',
    cval=0,
    zoom_range=0.2
)

def rgb2gray(rgb):
    if (len(rgb.shape) == 2):
        return rgb
    r, g, b = rgb[:,:,0], rgb[:,:,1], rgb[:,:,2]
    gray = 0.2989 * r + 0.5870 * g + 0.1140 * b

    return gray


def process_patch_full(im_np, y_np, file_id):
    if (y_np is None):
        y_np = np.zeros(im_np.shape[:2])
    else:
        if (len(y_np.shape) > 2):
            y_np = y_np[:,:,-1]
        
    y_np = y_np.astype(float)
    im_np = im_np.astype(float)
    
    im_np = rgb2gray(im_np)[:,:,np.newaxis]
    
    avg = np.mean(im_np)
    im_np -= avg
    im_np *= random.random() * 1.0 + 0.5
    im_np += avg + ((random.random() - 0.5) * 2.0) * 100
    
    
    y_np = y_np.astype(float)[:,:,np.newaxis]
    #y_np = (y_np * 2.0) - 1.0
    
    
    transform_params = datagen.get_random_transform(im_np.shape)
    for i in range(im_np.shape[2]):
        im_np[:,:,i] = datagen.apply_transform(im_np[:,:,i:i+1], transform_params)[:,:,0]
        
    y_np = datagen.apply_transform(y_np, transform_params)

    return im_np, y_np

def image_preprocessing(im_np):
    #im_np = scipy.misc.imresize(im_np, 0.25)
    im_np = rgb2gray(im_np)
    return im_np
